/*
 *  Copy me if you can.
 *  by 20h
 */

#include <u.h>
#include <libc.h>
#include <thread.h>
#include <mp.h>
#include <auth.h>
#include <libsec.h>

#define VERSION "ircc 3rd ed. Plan 9 4th ed."

int debug, ignoreflood, anickused, havenickserv, doautojoin;
char *lastchan, *clientinfo, *nick, *anick;
UserPasswd *passwd;

typedef struct command command;
struct command {
	char **params;
	int len;
	int isalloc;
};

command *ignorel, *joinl;

void
freecmd(command *cmd)
{
	int i;

	i = -1;
	if(cmd != nil) {
		if(cmd->isalloc)
			while(++i < cmd->len)
				if(cmd->params[i] != nil)
					free(cmd->params[i]);
		if(cmd->params != nil)
			free(cmd->params);
		free(cmd);
	}
	return;
}

int
addlist(command *c, char *p)
{
	int i;

	i = -1;
	if(c != nil) {
		if(c->isalloc) {
			p = strdup(p);
			while(++i < c->len) {
				if(c->params[i] == nil) {
					c->params[i] = p;
					return 2;
				}
			}
		}

		c->len++;
		c->params = realloc(c->params, sizeof(char *) * c->len);
		c->params[c->len - 1] = p;
		return 1;
	}

	return 0;
}

int
searchlist(command *c, char *p)
{
	int i;

	i = -1;
	if(c != nil)
		while(++i < c->len)
			if(c->params[i] != nil)
				if(!cistrcmp(c->params[i], p))
					return i;

	return -1;
}

int
dellist(command *c, char *p)
{
	int i;
	
	if(c != nil) {
		i = searchlist(c, p);
		if(i != -1) {
			free(c->params[i]);
			c->params[i] = nil;
			c->len--;
			return 1;
		}
	}

	return 0;
}

command *
parsecmd(command *cmd, char *data, int cmds)
{
	command *ret;
	char *next, *last, *temp;

	temp = nil;

	if(cmd == nil) {
		ret = mallocz(sizeof(command), 2);
		last = data;
	} else {
		ret = cmd;
		if(ret->isalloc) {
			temp = strdup(ret->params[ret->len - 1]);
			dellist(ret, ret->params[ret->len - 1]);
		} else {
			ret->len--;
			temp = ret->params[ret->len];
		}
		last = temp;
	}
	while((next = strchr(last, ' ')) != nil && ret->len < cmds) {
		*next = '\0';
		next++;
		if(last[0] == ':')
			last++;
		addlist(ret, last);
		last = next;
	}
	if(last[0] == ':')
		last++;
	if(last != nil)
		addlist(ret, last);

	if(temp != nil && cmd->isalloc)
		free(temp);

	return ret;
}

char *
mktimestamp(char bord, char bord_e)
{
	Tm *tim;
	char *ret;

	ret = malloc(31);
	tim = localtime(time(0));
	snprint(ret, 30, "%c%.2d:%.2d%c", bord, tim->hour, tim->min, bord_e);

	return ret;
}

void
setchan(char *new)
{
	if(lastchan != nil)
		free(lastchan);
	lastchan = mallocz(strlen(new) + 1, 2);
	strcpy(lastchan, new);
}

void
autojoin(int sock)
{
	int i;

	for(i = 0; i < joinl->len; i++)
		fprint(sock, "JOIN %s\r\n", joinl->params[i]);
	setchan(joinl->params[joinl->len - 1]);
}

int
gotcommand(int sock, char *data, int len)
{
	command *cmd;
	char *last, *next, *send, *cpy, *test, *tmstmp;
	int clen;

	clen = len;
	last = data;
	send = mallocz(513, 2);

	while((next = strchr(last, '\n')) !=  nil) {
		*next++ = '\0';
		clen -= (next - last);
		if(*(next - 2) == '\r')
			*(next - 2) = '\0';

		cpy = strdup(last);
		cmd = parsecmd(nil, last, 3);
		tmstmp = mktimestamp('(', ')');

		if(cmd->len > 1) {
			if(!cistrncmp(cmd->params[0], "PING", 4)) {
				fprint(sock, "PONG %s\r\n", cmd->params[1]);
				goto c_end;
			}
			if(!cistrncmp(cmd->params[0], "ERROR", 5)) {
				print("%sERROR%% %s %s %s\n", tmstmp, cmd->params[1], cmd->params[2], cmd->params[3]);
				goto c_end;
			}
		}
		if(cmd->len > 2) {
			if(!cistrncmp(cmd->params[1], "JOIN", 4)) {
				print("%s(%s) has joined %s\n", tmstmp, cmd->params[0], cmd->params[2]);
				goto c_end;
			}
			if(!cistrncmp(cmd->params[1], "PART", 4)) {
				print("%s(%s) has parted %s\n", tmstmp, cmd->params[0], cmd->params[2]);
				goto c_end;
			}
			if(!cistrncmp(cmd->params[1], "MODE", 4)) {
				print("%s(%s)-m-(%s)%% %s\n", tmstmp, cmd->params[0], cmd->params[2], cmd->params[3]);
				goto c_end;
			}
			if(!cistrncmp(cmd->params[1], "QUIT", 4)) {
				print("%s(%s) has quit (%s %s)\n", tmstmp, cmd->params[0], cmd->params[2], (cmd->len > 3) ? cmd->params[3] : "");
				goto c_end;
			}
			if((test = strchr(cmd->params[0], '!')) != nil)
				*test = '\0';
			if(!cistrncmp(cmd->params[1], "NICK", 4)) {
				print("%s(%s) changed nick to %s\n", tmstmp, cmd->params[0], cmd->params[2]);
				if(searchlist(ignorel, cmd->params[0]) != -1) {
					dellist(ignorel, cmd->params[0]);
					addlist(ignorel, cmd->params[2]);
				}
				goto c_end;
			}
		}
		if(cmd->len > 3) {
			if(strlen(cmd->params[1]) == 3) {
				if(ignoreflood) {
					if(!cistrncmp(cmd->params[1], "001", 3) ||
							!cistrncmp(cmd->params[1], "002", 3) ||
							!cistrncmp(cmd->params[1], "003", 3) ||
							!cistrncmp(cmd->params[1], "004", 3) ||
							!cistrncmp(cmd->params[1], "005", 3) ||
							!cistrncmp(cmd->params[1], "250", 3) ||
							!cistrncmp(cmd->params[1], "251", 3) ||
							!cistrncmp(cmd->params[1], "252", 3) ||
							!cistrncmp(cmd->params[1], "253", 3) ||
							!cistrncmp(cmd->params[1], "254", 3) ||
							!cistrncmp(cmd->params[1], "255", 3) ||
							!cistrncmp(cmd->params[1], "265", 3) ||
							!cistrncmp(cmd->params[1], "266", 3) ||
							!cistrncmp(cmd->params[1], "317", 3) ||
							!cistrncmp(cmd->params[1], "318", 3) ||
							!cistrncmp(cmd->params[1], "366", 3) ||
							!cistrncmp(cmd->params[1], "375", 3) ||
							!cistrncmp(cmd->params[1], "372", 3))
						goto c_end;
				}
				if(!cistrncmp(cmd->params[1], "376", 3)) {
					if(doautojoin && (!havenickserv || anickused))
						autojoin(sock);
					goto c_end;
				}
					
				print("%s(%s)(%s)%% %s\n", tmstmp, cmd->params[0], cmd->params[1], cmd->params[3]);
				if(!cistrncmp(cmd->params[1], "433", 3) && anick != nil) {
					if(anickused) {
						print("%s(IRCC)%% Sorry, but nick and anick are used. Try something different.\n", tmstmp);
						goto c_end;
					}
					fprint(sock, "NICK %s\r\n", anick);
					anickused = 1;
				}
				goto c_end;
			}
			if(!cistrncmp(cmd->params[1], "INVITE", 6)) {
				print("%s(%s) invited you to %s\n", tmstmp, cmd->params[0], cmd->params[3]);
				goto c_end;
			}
			if(!cistrncmp(cmd->params[1], "TOPIC", 5)) {
				print("%s(%s) topicchange in %s to \"%s\"\n", tmstmp, cmd->params[0], cmd->params[2], cmd->params[3]);
				goto c_end;
			}
			if(!cistrncmp(cmd->params[1], "KICK", 4)) {
				if(parsecmd(cmd, nil, 4) != nil)
					print("%s(%s) kicked %s out of %s \"%s\"\n", tmstmp, cmd->params[0], cmd->params[3], cmd->params[2], cmd->params[4]);
				else
					print("%s(%s) kicked %s out of %s\n", tmstmp, cmd->params[0], cmd->params[3], cmd->params[2]);
				goto c_end;
			}
			if(searchlist(ignorel, cmd->params[0]) == -1 &&
				searchlist(ignorel, cmd->params[2]) == -1) {
				if(!cistrncmp(cmd->params[1], "NOTICE", 6)) {
					print("%s(%s|%s)%% %s\n", tmstmp, cmd->params[0], cmd->params[2], cmd->params[3]);
					if(!cistrncmp(cmd->params[0], "NickServ", 8) && havenickserv) {
						print("in case NickServ\n");
						fprint(sock, "PRIVMSG NickServ :IDENTIFY %s\r\n", passwd->passwd);
						memset(passwd->passwd, 0, strlen(passwd->passwd));
						havenickserv = 0;
						if(doautojoin)
							autojoin(sock);
					}
					goto c_end;
				}
				if(!cistrncmp(cmd->params[1], "PRIVMSG", 7)) {
					if(*(cmd->params[3]) == '\x01') {
						if(!cistrncmp(cmd->params[3]+1, "VERSION", 7))
							fprint(sock, "NOTICE %s :\x01VERSION " VERSION "\x01\r\n", cmd->params[0]);
						if(!cistrncmp(cmd->params[3]+1, "TIME", 4)) {
							test = ctime(time(0));
							test[strlen(test) - 1] = '\0';
							fprint(sock, "NOTICE %s :\x01TIME %s\x01\r\n", cmd->params[0], test);
						}
						if(!cistrncmp(cmd->params[3]+1, "PING", 4))
							fprint(sock, "NOTICE %s :\x01PONG %s\r\n", cmd->params[0], cmd->params[3]+6);
						if(!cistrncmp(cmd->params[3]+1, "ACTION", 6)) {
							*(cmd->params[3] + strlen(cmd->params[3])-1) = '\0';
							print("%s(%s+%s) %s\n", tmstmp, cmd->params[0], cmd->params[2], cmd->params[3]+8);
							goto c_end;
						}
						if(!cistrncmp(cmd->params[3]+1, "CLIENTINFO", 10))
							fprint(sock, "NOTICE %s :\x01CLIENTINFO PING VERSION TIME USERINFO CLIENTINFO\x01\r\n", cmd->params[0]);
						if(!cistrncmp(cmd->params[3]+1, "USERINFO", 8))
							fprint(sock, "NOTICE %s :\x01USERINFO %s\x01\r\n", cmd->params[0], (clientinfo != nil) ? clientinfo : "<nil>");
					}
					print("%s(%s-%s)%% %s\n", tmstmp, cmd->params[0], cmd->params[2], cmd->params[3]);
					goto c_end;
				}
				if(!cistrncmp(cmd->params[0], "NOTICE", 6)) {
					print("%sNOTICE%% %s\n", tmstmp, cmd->params[3]);
					goto c_end;
				}
			} else
				goto c_end;
		}
		print("%s\n",cpy);
c_end:
		free(tmstmp);
		free(cpy);
		freecmd(cmd);
		last = next;
	}

	if(clen > 0)
		strcpy(data, data + len - clen);
	if(send != nil)
		free(send);

	return clen;
}

void
recvproc(void *sock)
{
	char *recv;
	int len, overl;

	recv = mallocz(1025, 2);
	len = 1;
	overl = 0;

	while(
#ifndef PLAN9PORT
		len >= 0 && 
#endif
#ifdef PLAN9PORT
		len > 0 &&
#endif
			*(int *)sock > 0) {
		memset(recv + overl, 0, 1025 - overl);
		len = read(*(int *)sock, recv + overl, 1024 - overl);

		if(debug)
			print("%s\n", recv);

		if(len > 1)
			overl = gotcommand(*(int *)sock, recv, len + overl);
	}

	free(recv);
	threadexitsall("recvproc");
}

void
usage(void)
{
	print("usage: %s [-dikt] [-a anick] [-c clientinfo] [-j #channel ...]"
			" [-n nick] [-p pass] [-u user] [net!]server[!port]\n", argv0);
	threadexitsall("usage");
}

void
threadmain(int argc, char *argv[])
{
	int sock, i, overl, tls;
	char *send, *rad, *tmstmp, *user, *pass;
	command *cmd;
	TLSconn *conn;

	user = "root";
	pass = nil;
	nick = "root";
	anick = nil;
	tls = 0;
	clientinfo = "The user has not specified his details.";
	joinl = mallocz(sizeof(command), 2);
	joinl->isalloc = 1;

	ARGBEGIN {
	case 'u':
		user = EARGF(usage());
		break;
	case 'p':
		pass = EARGF(usage());
		break;
	case 't':
		tls = 1;
		break;
	case 'd':
		debug = 1;
		break;
	case 'c':
		clientinfo = EARGF(usage());
		break;
	case 'i':
		ignoreflood = 1;
		break;
	case 'n':
		nick = EARGF(usage());
		break;
	case 'a':
		anick = EARGF(usage());
		break;
	case 'k':
		havenickserv = 1;
		break;
	case 'j':
		addlist(joinl, EARGF(usage()));
		doautojoin = 1;
		break;
	default:
		usage();
	} ARGEND;
			
			
	if(argc < 1)
		usage();

	if(havenickserv) {
		passwd = auth_getuserpasswd(auth_getkey, "proto=pass server=%s user=%s service=irc role=client", argv[0], nick);
		if(passwd == nil)
			sysfatal("auth_getuserpasswd: %r");
	}

	sock = dial(netmkaddr(argv[0], "tcp", tls ? "994" : "6667"), 0, 0, 0);
	if(sock < 0)
		sysfatal("dial: %r");

	if(tls) {
		conn = mallocz(sizeof(TLSconn), 2);
		overl = tlsClient(sock, conn);
		if(overl < 0)
			sysfatal("tlsClient: %r");
		sock = overl;

		if(conn->cert != nil)
			free(conn->cert);
	}

	if(user == nil)
		user = nick;

	if(pass != nil)
		fprint(sock, "PASS %s\r\n", pass);
	fprint(sock, "NICK %s\r\n", nick);
	fprint(sock, "USER %s localhost %s :%s\r\n", user, argv[0], user);

	if(proccreate(recvproc, &sock, 8192) < 0)
		sysfatal("procrfork: %r");

	ignorel = mallocz(sizeof(command), 2);
	ignorel->isalloc = 1;
	send = mallocz(513, 2);
	rad = mallocz(513, 2);

	while(sock > 0) {
		i = -1;
		cmd = nil;
		memset(send, 0, 513);
		memset(rad, 0, 513);

		while(read(0, &rad[++i], 1) && i < 400 && sock > 0) {
			if(rad[i] == '\n') {
				rad[i] = '\0';
				break;
			}
		}
		rad[i + 1] = '\0';

		tmstmp = mktimestamp('(', ')');
		if(rad[0] == '/')
			cmd = parsecmd(nil, rad, 2);

		if(cmd != nil) {
			if(cmd->len > 0) {
				switch(cmd->params[0][1]) {
				case 'H':
				case 'h':
					print("%s Help for ircc:\n", tmstmp);
					print("%s   /a chan text - send /me text to chan\n", tmstmp);
					print("%s   /c chan what - send CTCP what to chan\n", tmstmp);
					print("%s   /d - turn debugging on or off\n", tmstmp);
					print("%s   /h - show this help\n", tmstmp);
					print("%s   /i chan - ignore/unignore chan\n", tmstmp);
					print("%s   /j chan - join chan\n", tmstmp);
					print("%s   /k user [readon] - kick a user\n", tmstmp);
					print("%s   /l [chan] - list channels on server\n", tmstmp);
					print("%s   /m chan text - send text to chan\n", tmstmp);
					print("%s   /n nick - change nickname to nick\n", tmstmp);
					print("%s   /o chan mode [user] - set mode on chan\n", tmstmp);
					print("%s   /p chan - part chan\n", tmstmp);
					print("%s   /q [msg] - quit ircc\n", tmstmp);
					print("%s   /s chan - set active chan\n", tmstmp);
					print("%s   /t chan [topic] - set/show topic of chan\n", tmstmp);
					print("%s   /u chan - WHO of chan\n", tmstmp);
					print("%s   /w chan - WHOIS of chan\n", tmstmp);
					goto end_e;
				case 'D':
				case 'd':
					if(debug)
						debug = 0;
					else
						debug = 1;
					print("%s debug %d\n", tmstmp, debug);
					goto end_e;
				case 'Q':
				case 'q':
					sprint(send, "QUIT :\r\n");
					if(cmd->len == 2)
						sprint(send + 6, "%s\r\n", cmd->params[1]);
					if(cmd->len > 2)
						sprint(send + 6, "%s %s\r\n", cmd->params[1], cmd->params[2]);
					write(sock, send, strlen(send));
					close(sock);
					sock = -1;
					goto end_e;
				case 'L':
				case 'l':
					sprint(send, "LIST\r\n");
					if(cmd->len == 2)
						sprint(send + 4, " %s\r\n", cmd->params[1]);
					if(cmd->len > 2)
						sprint(send + 4, " %s %s\r\n", cmd->params[1], cmd->params[2]);
					write(sock, send, strlen(send));
					goto end_e;
				case 'S':
				case 's':
					if(cmd->len > 1)
						setchan(cmd->params[1]);
					else
						print("%s\n", (lastchan == nil) ? "<nil>": lastchan);
					goto end_e;
				default:
					break;
				}
			}
			if(cmd->len > 1) {
				switch(cmd->params[0][1]) {
				case 'J':
				case 'j':
					fprint(sock, "JOIN %s\r\n", cmd->params[1]);
					setchan(cmd->params[1]);
					goto end_e;
				case 'P':
				case 'p':
					fprint(sock, "PART %s\r\n", cmd->params[1]);
					goto end_e;
				case 'N':
				case 'n':
					fprint(sock, "NICK %s\r\n", cmd->params[1]);
					goto end_e;
				case 'W':
				case 'w':
					fprint(sock, "WHOIS %s\r\n", cmd->params[1]);
					goto end_e;
				case 'U':
				case 'u':
					fprint(sock, "WHO %s\r\n", cmd->params[1]);
					goto end_e;
				case 'T':
				case 't':
					sprint(send, "TOPIC %s\r\n", cmd->params[1]);
					if(cmd->len > 2)
						sprint(send + strlen(send) - 2, " :%s\r\n", cmd->params[2]);
					write(sock, send, strlen(send));
					goto end_e;
				case 'I':
				case 'i':
					if(searchlist(ignorel, cmd->params[1]) != -1) {
						dellist(ignorel, cmd->params[1]);
						print("%s no more ignored. %d\n", cmd->params[1], searchlist(ignorel, cmd->params[1]));
					} else {
						addlist(ignorel, cmd->params[1]);
						print("%s will be ignored.\n", cmd->params[1]);
					}
					goto end_e;
				default:
					break;
				}
			}
			if(cmd->len > 2) {
				switch(cmd->params[0][1]) {
				case 'M':
				case 'm':
					fprint(sock, "PRIVMSG %s :%s\r\n", cmd->params[1], cmd->params[2]);
					goto end_e;
				case 'A':
				case 'a':
					fprint(sock, "PRIVMSG %s :\x01ACTION %s\x01\r\n", cmd->params[1], cmd->params[2]);
					goto end_e;
				case 'O':
				case 'o':
					if(cmd->len > 3)
						fprint(sock, "MODE %s %s %s\r\n", cmd->params[1], cmd->params[2], cmd->params[3]);
					else
						fprint(sock, "MODE %s %s\r\n", cmd->params[1], cmd->params[2]);
					goto end_e;
				case 'C':
				case 'c':
					print("%s(%s) ctcp %s\n", tmstmp, cmd->params[1], cmd->params[2]);
					fprint(sock, "PRIVMSG %s :\x01%s\x01\r\n", cmd->params[1], cmd->params[2]);
					goto end_e;
				case 'K':
				case 'k':
					sprint(send, "KICK %s %s\r\n", cmd->params[1], cmd->params[2]);
					if(cmd->len > 3)
						sprint(send + strlen(send) - 2, " :%s\r\n", cmd->params[3]);
					write(sock, send, strlen(send));
					goto end_e;
				default:
					break;
				}
			}
		}
		if(lastchan != nil)
			fprint(sock, "PRIVMSG %s :%s\r\n", lastchan, rad);
		else
			print(".");
end_e:
		print("%s\n", tmstmp);
		free(tmstmp);
		freecmd(cmd);
	}

	if(lastchan != nil)
		free(lastchan);
	free(send);
	free(rad);
	freecmd(ignorel);
	freecmd(joinl);

	threadexitsall(0);
}
